#!/usr/bin/env python

from collections import defaultdict
import glob
import os

import numpy as np
import pandas
import xarray

import smartcarb.io
import smartcarb.misc

from . import timeprofiles as tp




def filter_results(ts):
    """
    Remove bad fits of flux inversion.
    """
    tracer = 'CO2' if 'CO2 emissions (Mt/yr)' in ts else 'NO2'

    if tracer == 'CO2':
        good = np.isfinite(ts['CO2 emissions (Mt/yr)'])
        good = ts['CO2 emissions (Mt/yr)'] > 0.0
        good &= ts['CO2 uncertainty (Mt/yr)'] < 10.0

        return (
            ts['CO2 emissions (Mt/yr)'][good],
            ts['CO2 uncertainty (Mt/yr)'][good],
            ts.lon_eq[good],
            good
        )
    else:
        good = np.isfinite(ts['NO2 emissions (kt/yr)'])
        good &= ts['NO2 emissions (kt/yr)'] > 0
        good &= ts['NO2 uncertainty (kt/yr)'] < 10.0

        return (
            ts['NO2 emissions (kt/yr)'][good],
            ts['NO2 uncertainty (kt/yr)'][good],
            ts.lon_eq[good],
            good
        )



def read_results(filename, nsat=1):
    ts = pandas.read_csv(filename, delimiter='\t', index_col=0,
                         parse_dates=True, na_values='%10.3f' % np.nan)
    ts = smartcarb.misc.filter_dataset(ts, nsat)

    return ts


def read_analytical_inversion(filename, nsat=1, use_constant=False, origin='Berlin',
            tracer='CO2', fit_cycle=True, n_knots=4):
    """
    Read and prepare time series of estimated and true emissions as well as the
    fitted seasonal cycle.
    """
    ts = read_results(filename, nsat)
    lon_eqs_in_constellation = smartcarb.misc.lon_eq_by_nsat(nsat)

    em_true = smartcarb.io.read_emissions(origin=origin, time='raw',
                                          use_constant=use_constant,
                                          tracer=tracer)
    if tracer == 'NO2':
        em_true *= 1e3
        units = 'kt NO2 yr-1'
    else:
        units = 'Mt CO2 yr-1'

    # remove bad values
    est, est_std, lon_eq, goods = filter_results(ts)

    # create xarray dataset with all data
    d = xarray.Dataset()

    d['overpass'] = xarray.DataArray(est.index, dims='overpass')
    d['est_emissions'] = xarray.DataArray(est, dims='overpass')
    d['est_emissions_std'] = xarray.DataArray(est_std, dims='overpass')
    d['lon_eq'] = xarray.DataArray(lon_eq, dims='overpass')
    d['orbit'] = xarray.DataArray(ts['orbit'].loc[est.index], dims='overpass')


    d['time'] = xarray.DataArray(em_true.index, dims='time')
    d['true_emissions'] = xarray.DataArray(em_true.values, dims='time')

    if fit_cycle:
        # fit seasonal cycle
        fit, func, times, cycle, chi2 = tp.fit(d['overpass'], d['est_emissions'],
                                               d['est_emissions_std'], n_knots=n_knots)

        d['fitted_cycle'] = xarray.DataArray(cycle, dims='time')
        d['fitted_cycle'].attrs['annual_mean'] = func.integrate(fit.x)
        d['fitted_cycle'].attrs['units'] = units

    return d



def get_output_path(co2_ssp, no2_ssp):
    """
    Return output path for mass-balance approach based on single sounding
    precision of CO2 and NO2 instrument.
    """
    output_path = '/output/SMARTCARB/output/mass-balance/'

    if no2_ssp == 'none':
        return os.path.join(output_path, 'CO2-%s' % co2_ssp)
    else:
        return os.path.join(output_path, 'CO2-%s_and_NO2-%s' %
                                   (co2_ssp, no2_ssp))



def read_background_errors(tracer, co2_ssp='medium', no2_ssp='high'):
    """
    """
    # add results from mass-balance approach
    data = defaultdict(list)
    output_path = get_output_path(co2_ssp, no2_ssp)
    filenames = sorted(glob.glob(
        os.path.join(output_path, "*_data.nc")
    ))
    results = defaultdict(list)

    for filename in filenames:

        data = xarray.open_dataset(filename)

        if 'detected_plume' not in data:
            continue

        time = data.attrs['time']
        bias = data['%s_estimated_background' % tracer].attrs['bias']
        scatter = data['%s_estimated_background' % tracer].attrs['scatter']

        data.close()

        if np.isnan(bias) and np.isnan(scatter):
            continue

        results['overpass'].append(time)
        results['bias'].append(bias)
        results['scatter'].append(scatter)


    data = xarray.Dataset()
    data['overpass'] = xarray.DataArray(results['overpass'], dims='overpass')
    data['bias'] = xarray.DataArray(results['bias'], dims='overpass')
    data['scatter'] = xarray.DataArray(results['scatter'], dims='overpass')

    return data


def read_mass_balance(tracer, which='combined', method='sub-areas',
                      co2_ssp='medium', no2_ssp='high', wind_threshold=0.0):


    bad_times = []
    if no2_ssp == 'high':
        bad_times = [
            np.datetime64('2015-05-22T11:00:00'), # very cloudy and only
                                                  # polygon at the end of 
                                                  # the deteteced plume
            np.datetime64('2015-06-11T11:00:00'), # wind direction turns
            np.datetime64('2015-07-03 11:00'), # janschwalde in plume
            np.datetime64('2015-08-20 11:00'), # janschwalde in plume
            np.datetime64('2015-10-01 11:00'), # janschwalde in plume
        ]
    else:
        bad_times = [
            np.datetime64('2015-08-20 11:00'), # janschwalde in plume
        ]

    # add true emissions
    em_true = smartcarb.io.read_emissions(origin='Berlin', time='raw',
                                          use_constant=False,
                                          tracer=tracer)

    if tracer == 'NO2':
        em_true *= 1e3 # Mt yr-1 -> kt yr-1

    dataset = xarray.Dataset()
    dataset['time'] = xarray.DataArray(em_true.index, dims='time')
    dataset['true_emissions'] = xarray.DataArray(em_true.values, dims='time')

    # add results from mass-balance approach
    data = defaultdict(list)
    output_path = get_output_path(co2_ssp, no2_ssp)
    filenames = sorted(glob.glob(
        os.path.join(output_path, "*_%s_%s-results.nc") % (method, tracer)
    ))

    varname = {
        'model':      'flux_from_model_tracer_using_true_wind_fit',
        'combined':   'flux_from_%s_minus_estimated_background_using_mean_wind_fit' % tracer,
        'background': 'flux_from_%s_noisefree_minus_estimated_background_using_effective_wind_fit' % tracer,
        'wind':       'flux_from_%s_true_signal_using_mean_wind_fit' % tracer,
        'retrieval':  'flux_from_%s_minus_true_background_using_effective_wind_fit' % tracer,
        'method':     'flux_from_%s_true_signal_using_effective_wind_fit' % tracer
    }[which]

    for filename in filenames:
        d = xarray.open_dataset(filename)

        if pandas.Timestamp(d.attrs['time']) in bad_times:
            print('Skip bad overpass')
            continue

        if d.attrs['wind_speed_500m'] < wind_threshold:
            print('Skip wind speed below %.1f m/s' % wind_threshold)
            continue

        data['overpass'].append( pandas.Timestamp(d.attrs['time']) )
        data['orbit'].append( d.attrs['orbit'] )
        data['lon_eq'].append( d.attrs['lon_eq'] )


        if d[varname].emissions < 0 or d[varname].emissions > 100:
            print('Est. emission of %.1f set to nan. ' % d[varname].emissions)
            data['est_emissions'].append(np.nan)
            data['est_emissions_std'].append(np.nan)
        else:
            data['est_emissions'].append(d[varname].emissions)
            data['est_emissions_std'].append(d[varname].emissions_std)

        # wind speed and direction
        data['wind_speed_500m'].append( d.attrs['wind_speed_500m'] )
        data['wind_direction_500m'].append( d.attrs['wind_direction_500m'] )
        data['wind_speed_eff'].append( d.attrs['wind_speed_eff'] )

        d.close()

    for key, values in data.items():
        dataset[key] = xarray.DataArray(values, dims='overpass')

    return dataset



def read_old_mass_balance():
    path = '/project/kug/smartcarb/results/Berlin/flux_inversion/mass-balance/results'

    ts = defaultdict(list)

    for fn in sorted(glob.glob("*NO2-high_and_0.5_results.dat")):
        with open(fn) as res:
            for line in res:
                key, value = line.strip().split('\t')

                if key == 'time':
                    value = np.datetime64(value)
                else:
                    try:
                        value = float(value)
                    except ValueError:
                        pass

                ts[key].append(value)

    ts = pandas.DataFrame(ts)
    ts = smartcarb.misc.filter_dataset(ts, nsat='6')

    data = xarray.Dataset()
    data['overpass'] = xarray.DataArray(ts['time'], name='overpass', dims='overpass')

    data['est_emissions'] = xarray.DataArray(ts['est. CO2 emissions (Mt/yr)'].values, dims='overpass')
    data['est_emissions_std'] = xarray.DataArray(ts['est. CO2 emissions uncertainty (Mt/yr)'].values,
                                                 dims='overpass')

    data['lon_eq'] = xarray.DataArray(ts['lon_eq'].values, dims='overpass')
    data['orbit'] = xarray.DataArray(ts['orbit'].values, dims='overpass')

    # add true emissions
    em_true = smartcarb.io.read_emissions(origin='Berlin', time='raw',
                                          use_constant=False,
                                          tracer='CO2')

    data['time'] = xarray.DataArray(em_true.index, dims='time')
    data['true_emissions'] = xarray.DataArray(em_true.values, dims='time')

    return data


def read_all():
    ts = defaultdict(dict)

    # analytical inversion
    data_path = "/project/kug/smartcarb/results/Berlin/flux_inversion/analytical/"
    fn = {
        'low':    os.path.join(data_path, "CO2_Berlin_vari_model-bg_05ppm.dat"),
        'medium': os.path.join(data_path, "CO2_Berlin_vari_model-bg_07ppm.dat"),
        'high':   os.path.join(data_path, "CO2_Berlin_vari_model-bg_10ppm.dat")
    }

    for key, filename in fn.items():
        ts['ana-vari'][key] = read_analytical_inversion(filename, '6', tracer='CO2',
                                                        use_constant=False)

        filename = filename.replace('vari', 'constant')
        ts['ana-const'][key] = read_analytical_inversion(filename, '6', tracer='CO2',
                                                         use_constant=True)

    # CO2- and NO2-based plume detection
    for method in ['sub-areas', 'gauss']:
        for noise in ['low', 'medium', 'high']:

            # CO2-based plume detection
            ts['co2-%s' % method][noise] = read_mass_balance(
                'CO2', which='combined', method=method, co2_ssp=noise,
                no2_ssp='none', wind_threshold=0.0
            )

            # NO2-based plume detection
            ts['no2-%s' % method][noise] = read_mass_balance(
                'CO2', which='combined', method=method, co2_ssp=noise,
                no2_ssp='high', wind_threshold=0.0
            )

    return ts








