
import itertools

from matplotlib.dates import DateFormatter
import matplotlib.pyplot as plt
import numpy as np
import pandas
import scipy.stats

import smartcarb.misc
from amrs.misc.chem import ppm_to_mass, kgs_to_Mtyr

from . import timeprofiles as tp
from . import io


COLORS = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
          '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']

LON_EQ_TO_LETTER = {
    0: 'a',  401: 'b', 805: 'c', 1207: 'd', 1610: 'e', 2012: 'f'
}

def extract_time_range(time, values, tmin, tmax):

    data = []

    for hour in range(tmin, tmax+1):
        mask = np.array([pandas.Timestamp(t).hour == hour for t in
                         np.array(time)])

        data.append( values[mask] )

        new_time = np.array(time)[mask]


    return new_time, np.array(data)




def init_mass_balance_plot(tracer, distances, delta, emissions, q_true, u_1km):

    xvalues = (distances[1:] - delta / 2) / 1e3 # m -> km

    fig = plt.figure()

    # [left, bottom, width, height]
    ax =  fig.add_axes([0.12, 0.37, 0.85, 0.57])
    ax2 = fig.add_axes([0.12, 0.10, 0.85, 0.23])

    lines = []


    lines.append(
        ax.fill_betweenx([-20,80], [-20,-20], [+20,+20], color='#eeeeee',
                          label='city area')
    )
    lines.append(
        ax2.fill_betweenx([0,80], [-20,-20], [+20,+20], color='#eeeeee',
                          label='_none_')
    )

    lines += ax.plot([xvalues[0]-delta/2e3, xvalues[-1]+delta/2e3],
                     [q_true, q_true], 'k-', label='true emissions (10:30 UTC)')

    label = 'range of true emissions (6-12 UTC)'
    lines.append(
        ax.fill_between([xvalues[0]-delta/2e3, xvalues[-1]+delta/2e3],
                        [min(emissions), min(emissions)],
                        [max(emissions), max(emissions)],
                        label=label, color='silver')
    )

    # constant wind speed
    lines += ax2.plot(xvalues, np.full(xvalues.shape, u_1km), 's-',
                       label='wind speed over Berlin (0 - 500 m)')

    ax2.set_xlabel('along plume distance from source (km)')

    if tracer == 'CO2':
        ax.set_ylabel('CO$_2$ flux (Mt CO$_2$ yr$^{-1}$)')
    else:
        ax.set_ylabel('NO$_x$ flux (kt NO$_2$ yr$^{-1}$)')

    return ax, ax2, lines



def add_mass_balance_plot(i, ax, ax2, tracer, distances, delta, flux, flux_std,
                          ueff, lines=None, add_wind=False,
                          dq=np.nan, fit=None):

    if lines is None:
        lines = []

    xvalues = (distances[1:] - delta / 2) / 1e3 # m -> km

    if add_wind:
        lines += ax2.plot(xvalues, ueff, 'o-',
                     label='wind speed in plume')

    factor = 1.0 if  tracer == 'CO2' else 1e3

    label = [
        'model tracer + winds',
        'method',
        'wind',
        'instrument noise',
        'background',
        'combined errors',
    ][i]
    marker = ['D', 'o', 's', 'd', 'v', '^'][i]

    #extra = '$\Delta$Q = %+.1f %st yr$^{-1}$' % (dq, 'k' if tracer ==
    #                                     'NO2' else 'M')
    #label = labels[i] % extra

    if np.any(np.isnan(flux_std)):
        lines += ax.errorbar(xvalues, factor * kgs_to_Mtyr(flux),
                             ls='', marker=marker,
                             label=label, color=COLORS[i])

    else:
        lines.append( ax.errorbar(xvalues, factor * kgs_to_Mtyr(flux),
                             factor * kgs_to_Mtyr(flux_std),
                             ls='', marker=marker,
                             label=label, color=COLORS[i])
                    )

    ax.plot(xvalues, fit, label='__none__',
            color=COLORS[i])
    ax.set_xticklabels([])

    ax.set_xlim(min(distances)/1e3, max(distances)/1e3)
    ax2.set_xlim(min(distances)/1e3, max(distances)/1e3)

    return lines



def finalize_mass_balance_plot(ax, ax2, tracer, lines):

    ax.set_ylim(bottom=-20, top=80 if tracer == 'NO2' else 80)

    ax2.set_ylabel('wind speed (m s$^{-1}$)')
    ax2.set_ylim(0, 20)

    ax.set_zorder(ax2.get_zorder()+1)
    ax.patch.set_visible(False)

    ax.legend(loc='upper left', fontsize='small', ncol=2)
    ax2.legend(loc='upper left', fontsize='small')

    ax.grid(True)
    ax2.grid(True)


def compute_bias_and_scatter(ts, tracer, n_knots):
    bias = []
    scatter = []
    for i in [1,2,3,4,5]:
        diffs, mb, sd = plot_annual_emissions_by_constellation(ts,
                                                               n_knots=n_knots,
                                                               tracer=tracer,
                                                               constellation=i,
                                                               show=False,
                                                               gamma=None)
        bias.append(mb)
        scatter.append(sd)

    return np.array(bias), np.array(scatter)


def plot_bias_and_scatter_for_number_of_satellites(ts_analy, ts_mb,
                                                   tracer='CO2', n_knots=3):
    """
    """
    bias_anal, scatter_anal = compute_bias_and_scatter(ts_analy, tracer,
                                                       n_knots)
    bias_mb, scatter_mb = compute_bias_and_scatter(ts_mb, tracer, n_knots)

    fig, axes = plt.subplots(1, 2)
    nsats = np.arange(1,6)

    axes[0].plot(nsats, 100 * np.array(bias_anal), 'o--',
                 label='analytical inversion')

    axes[0].plot(nsats, 100 * np.array(bias_mb), 's--',
                 label='mass-balance')

    axes[1].plot(nsats, 100 * np.array(scatter_anal), 'o--',
                 label='analytical inversion')

    axes[1].plot(nsats, 100 * np.array(scatter_mb), 's--',
                 label='mass-balance')

    axes[0].set_ylabel('bias (%)')
    axes[1].set_ylabel('scatter (%)')

    for i in [0,1]:
        axes[i].grid(True)
        axes[i].set_xlabel('number of satellites')
        axes[i].legend()
        axes[i].set_xlim(0.5, 5.5)
        axes[i].set_ylim(0, 35)

        axes[i].set_xticks(nsats)

    plt.tight_layout()




def plot_annual_emissions_by_constellation(data, n_knots=1, tracer='CO2',
                                           constellation='eq', show=True,
                                           gamma=None):

    # constellations
    if constellation == 'eq': # equidistant spacing
        nsats = ['a', 'b', 'c', 'd', 'e', 'f', # TODO:use 'abc' etc.
                 'ad', 'be', 'cf', 'ace', 'bdf', 'abcdef']

        steps = [(1,0,6), (2,6,9), (3,9,11), (6,11,12)]
    else:
        nsats = [''.join(c) for c in itertools.combinations('abcdef', constellation)]
        steps = [(constellation,0,len(nsats))]

    annual_emissions = []
    annual_emissions_std = []
    overpasses = []

    for nsat in nsats:
        cycle = tp.add_cycle(data, n_knots=n_knots, nsat=nsat, gamma=gamma)
        em = cycle['fitted_cycle'].attrs['annual_mean']
        em_std = cycle['fitted_cycle'].attrs['annual_mean_std']
        annual_emissions.append(em)
        annual_emissions_std.append(em_std)

        overpasses.append(
            cycle['fitted_cycle'].attrs['number of overpasses']
        )

    annual_emissions = np.array(annual_emissions)
    annual_emissions_std = np.array(annual_emissions_std)
    overpasses = np.array(overpasses)

    # true emissions
    true_emissions = {'CO2': 20.0, 'NO2': 22.1}[tracer]


    #
    mask = annual_emissions_std > true_emissions
    if np.any(mask):
        print('Est uncertainty larger than true emissions')
        annual_emissions[mask] = np.nan
        annual_emissions_std[mask] = np.nan


    differences = []
    bias = []
    scatter = []

    if show:
        plt.figure()

    for nsat, i, j in steps:

        if nsat == 6:
            delta = np.nan
            label = '#sats = %d' % nsat
        else:
            delta = np.nanmax(annual_emissions[i:j]) - np.nanmin(annual_emissions[i:j])
            delta /= 20.0 if tracer == 'CO2' else 22.1
            label = '#sats = %d ($\Delta$ = %d%%)' % (nsat, round(100*delta))

        if show:
            label = '#sats = %d' % nsat
            plt.bar(np.arange(i,j), annual_emissions[i:j], tick_label=nsats[i:j],
                    label=label)

            plt.errorbar(np.arange(i,j), annual_emissions[i:j], annual_emissions_std[i:j],
                         ls='', marker='', color='k')

            nans = np.isnan(annual_emissions[i:j])
            plt.plot(np.arange(i,j)[nans], np.full(nans.sum(), 2.5), 'kx', ms=15)

        difference = annual_emissions[i:j] - true_emissions
        mb = np.nanmean(difference) / true_emissions
        sd = np.nanstd(difference - true_emissions) / true_emissions

        differences.append(difference)
        bias.append(mb)
        scatter.append(sd)


    if show:
        if tracer == 'CO2':
            plt.plot([-0.5, len(nsats)-0.5], [20.0, 20.0], 'k-', label='true emissions')
            plt.ylabel('CO$_2$ emissions (Mt yr$^{-1}$)')
        else:
            plt.plot([-0.5, 11.5], [22.1, 22.1], 'k-', label='true emissions')
            plt.ylabel('NO$_x$ emissions (kt yr$^{-1}$)')

        plt.xlabel('constellations')

        plt.xticks(np.arange(len(nsats)), nsats, rotation=45)

        plt.legend(ncol=2)

        plt.ylim(0, 50)

        for i, o in enumerate(overpasses):
            if np.isnan(annual_emissions[i]):
                y = 5.0
            else:
                y = annual_emissions[i] + annual_emissions_std[i]

            plt.text(i, y + 0.5,
                     '(%s)' % o, fontsize='small',
                     ha='center')



        plt.tight_layout()


    return np.array(differences), np.array(mb), np.array(sd)



def plot_ts(tracer, data, cycle=None, nsat=1, label=None, origin='Berlin',
            n_knots=3, gamma=None):

    # plot results
    if tracer == 'NO2':
        units = 'kt yr$^{-1}$'
    elif tracer == 'CO2':
        units = 'Mt yr$^{-1}$'
    else:
        units = 'ratio'


    if tracer == 'ratio':
        var_true = 'true_ratio'
        var_est = 'est_ratio'
        var_std = 'est_ratio_std'
    else:
        var_true = 'true_emissions'
        var_est = 'est_emissions'
        var_std = 'est_emissions_std'

    fig = plt.figure(figsize=(8.4, 3.5))
    ax = fig.add_subplot(111)


    em_date, em_range = extract_time_range(data['time'],
                                           data[var_true], 10, 11)


    ax.plot(em_date, em_range.mean(0), 'k-',
            label='true emissions (10-11 UTC: %.1f %s)' % (em_range.mean(),
                                                           units))

    ax.fill_between(em_date, em_range.min(0), em_range.max(0), color='silver')


    lon_eqs_in_constellation = smartcarb.misc.lon_eq_by_nsat(nsat)

    # statistics
    valids = np.isin(data.lon_eq, lon_eqs_in_constellation)
    true_values = data[var_true].loc[data.overpass].values[valids]
    est_values = data[var_est].values[valids]

    mb = np.nanmean(est_values - true_values)
    sd = np.nanstd(est_values - true_values)
    rmsd = np.sqrt(mb**2 + sd**2 / sd.size)

    print(mb, sd, rmsd)


    # fit seasonal cycle
    tp.add_cycle(data, n_knots=n_knots, nsat=nsat, gamma=gamma)


    for i, lon_eq in enumerate(lon_eqs_in_constellation, 1):

        letter = LON_EQ_TO_LETTER[lon_eq]

        sat = data['lon_eq'] == lon_eq

        times = data['overpass'].loc[sat].values
        values = data[var_est].loc[sat].values
        errors = data[var_std].loc[sat].values
        #errors = np.sqrt(data['fitted_cycle'].attrs['chi2'])

        ax.errorbar(times, values, errors, ls='', marker='o',
                    label='#%s (%d plumes)' % (letter, sum(sat)))

    if label is not None:
        for i,(t,v) in enumerate(zip(times,values)):
            s = '%.2g' % ts[label][i]
            ax.text(t,v,s)

    ax.xaxis.set_major_formatter(DateFormatter('%b'))

    ax.set_xlim('2015-01-01', '2015-12-31')
    if origin == 'Berlin':
        if tracer == 'ratio':
            ax.set_ylim(200,1800)
        else:
            ax.set_ylim(0,max(50, np.nanmax(values) ))
    else:
        ax.set_ylim(ymin=0)

    ax.set_xlabel('2015')

    if tracer == 'CO2':
        ax.set_ylabel('CO$_2$ emissions (%s)' % units)
    elif tracer == 'NO2':
        ax.set_ylabel('NO$_2$ emissions (%s)' % units)
    else:
        ax.set_ylabel('CO$_2$-to-NO$_2$ emission ratio')

    # plot cycle
    if tracer != 'ratio' and 'fitted_cycle' in data:
        ax.plot(data['time'], data['fitted_cycle'], 'r-',
                label='fitted seasonal cycle (%.1f$\pm$%.1f %s)' %
                (data['fitted_cycle'].attrs['annual_mean'],
                 data['fitted_cycle'].attrs['annual_mean_std'],
                 units))

    if tracer == 'ratio' and 'fitted_ratio_cycle' in data:
        ax.plot(data['time'], data['fitted_ratio_cycle'], 'r-',
                label='fitted seasonal cycle (%.1f$\pm$%.1f %s)' %
                (data['fitted_ratio_cycle'].attrs['annual_mean'],
                 data['fitted_ratio_cycle'].attrs['annual_mean_std'],
                 units))


    ax.grid(True, ls='-', color='gray')

    if isinstance(nsat, str):
        nsat = int(nsat[:1])

    ncols = {
        1: 1,
        2: 3,
        3: 3,
        4: 3,
        5: 3,
        6: 3}[nsat]

    plt.legend(loc=1, ncol=ncols, fontsize='small')
    plt.tight_layout(pad=0.25)
    fig.subplots_adjust(top=0.95)



def filter_nsat(d, nsat):
    lon_eqs_in_constellation = smartcarb.misc.lon_eq_by_nsat(nsat)
    valids = np.isin(d.lon_eq, lon_eqs_in_constellation)
    return d.loc[{'overpass': d.overpass.values[valids]}]


def plot_mb_histograms(against='method', method='sub-areas'):

    ts_method = io.read_mass_balance('CO2', which='method', method=method)

    # bias and scatter of method
    d = (ts_method.est_emissions - ts_method.true_emissions.loc[ts_method.overpass]).values
    mb = np.nanmean(d)
    sd = np.nanstd(d)
    n = np.isfinite(d).size

    print('Method:')
    print('- MB:', mb, 100 * mb / 20.0)
    print('- SD:', sd, 100 * sd / 20.0)
    print('- N :', n)

    fig, axes = plt.subplots(ncols=4, figsize=(12, 3.0))

    for i, var in enumerate(['retrieval', 'background', 'wind', 'combined']):

        ts = io.read_mass_balance('CO2', which=var, method=method)



        if against == 'method':
            d = (ts - ts_method).est_emissions.values
        else:
            d = (ts.est_emissions - ts_method.true_emissions.loc[ts.overpass]).values

        mb = np.nanmean(d)
        sd = np.nanstd(d)
        n = np.isfinite(d).size

        text = 'MB: %.1f Mt yr$^{-1}$ (%d %%)' % (mb, 100 * mb / 20.0)
        axes[i].text(-29.2, 12.4, text, va='top', fontsize='small')

        text = 'SD: %.1f Mt yr$^{-1}$ (%d %%)' % (sd, 100 * sd / 20.0)
        axes[i].text(-29.2, 11.2, text, va='top', fontsize='small')

        text = 'N: %d' % n
        axes[i].text(-29.2, 9.8, text, va='top', fontsize='small')

        axes[i].hist(d, bins=np.linspace(-30,30,31))
        axes[i].set_xlabel('$\Delta Q$ (Mt yr$^{-1}$)')

        axes[i].set_xlim(-30,30)
        axes[i].set_xticks(np.arange(-30,31,10))
        axes[i].set_ylim(0, 12.5)

        axes[i].set_title('(%s) %s' % ('abcde'[i], var))

        if i == 0:
            axes[i].set_ylabel('number of estimates')


    plt.tight_layout()






def boxplot(datasets, nsat='3a', nmax=None, sample='all', indx=None,
            origin='Berlin', use_constant=False, vmin=-20, vmax=20,
            is_mce=False):
    """
    """

    # filter fo nsat
    datasets = [filter_nsat(d, nsat) for d in datasets]

    data = []
    unc = []

    for d in datasets:
        y = d['est_emissions'] - d['true_emissions'].loc[d.overpass]
        u = d['est_emissions_std']

        valids = Ellipsis # np.isfinite(y) & np.isfinite(u)

        data.append(y[valids])
        unc.append(u[valids])


    if sample == 'same':
        ind = sorted(set(datasets[0].overpass.values) &
                     set(datasets[1].overpass.values) &
                     set(datasets[2].overpass.values))
        nn = [len(v) for v in data]
        data = [d.loc[ind] for d in data]
    else:
        nn = None

    if indx is not None:
        data = [d[indx] for d in data]

    sep = ' & '
    line = sep.join(['noise', '#plumes', 'MB', 'MB', 'SD', 'SD',
                      'RMSD', 'RMSD', 'unc', 'unc'])
    print(line)

    for noise, d,u in zip([0.5, 0.7, 1.0], data, unc):

        # annual mean emissions at overpass used for computing relative errors
        if origin == 'Berlin':
            t = 100 / 20.0 # em_mean.mean()
        elif origin == 'Janschwalde':
            t = 100 / 40.0 # Janschwalde (at overpass)
        else:
            raise ValueError

        n = len(d)
        mb  = np.mean(d)
        std = np.std(d)
        rms = np.sqrt(mb**2 + std**2 / n)

        print('%.1f' % noise, end=sep)
        print('%5d' % n, end=sep)
        print('%5.1f%s%5.1f' % (mb, sep, mb*t), end=sep)
        print('%5.1f%s%5.1f' % (std, sep, std*t), end=sep)
        print('%5.2f%s%5.1f' % (rms, sep, rms*t), end=sep)
        print('%5.1f%s%5.1f' % (np.mean(u), sep, np.mean(u)*t))

    print('Mean emissions: %.2f Mt/yr' % 20.0) #em_mean.mean())

    fig = plt.figure(figsize=(3.3,3.5))
    ax = fig.add_subplot(111)

    for i,n in enumerate(data):
        if nn is None:
            ax.text(i+1,vmax+0.1, str(len(n)), ha='center', va='bottom')
        else:
            text = '%s/%s' % (len(n), nn[i])
            ax.text(i+1,vmax+0.1, text, ha='center', va='bottom')


    ax.boxplot(data, showfliers=True, whis=[5,95], bootstrap=5000, widths=0.4,
               labels=['0.5', '0.7', '1.0'], meanline=True, showmeans=True)

    ax.set_xlabel('$\sigma_{VEG50}$ (ppm)')
    ax.set_ylabel('emission estimation error (Mt yr$^{-1}$)')

    ax.set_ylim(vmin,vmax)
    ax.grid(True, ls='-', color='gray')
    plt.tight_layout(pad=0.25)
    fig.subplots_adjust(top=0.95)

    return data




def plot_wind_error():
    """
    Plot two plots showing wind speed and wind error (difference between
    500 m and eff wind speed)
    """
    fig, axes = plt.subplots(nrows=2, figsize=(9,4))

    ts_method = io.read_mass_balance('CO2', which='method',
                                     method='sub-areas')
    ts_wind = io.read_mass_balance('CO2', which='wind',
                                   method='sub-areas')

    axes[0].plot(ts_wind.overpass, ts_wind.wind_speed_eff, 'o', mfc='none',
                 label='effective wind speed')

    axes[0].plot(ts_wind.overpass, ts_wind.wind_speed_500m, 's', mfc='none',
                 label='1-km wind speed')


    axes[0].set_title('(a)')
    axes[0].set_ylim(0,15)
    axes[0].set_ylabel('wind speed (m s$^{-1}$)')
    axes[0].set_xlim('2015-01-01', '2016-01-01')
    axes[0].grid(True)
    axes[0].legend(ncol=2)



    d_wind = ts_wind.wind_speed_500m - ts_method.wind_speed_eff
    d_emis = ts_wind.est_emissions - ts_method.est_emissions

    valids = np.isfinite(d_wind) & np.isfinite(d_emis)
    d_wind = d_wind[valids]
    d_emis = d_emis[valids]


    lines = axes[1].plot(d_wind.overpass, d_wind, 'o',
                         label='wind', mfc='none')

    axes[1].set_title('(b)')
    axes[1].set_ylim(-8, 8)
    axes[1].set_ylabel('wind error (m s$^{-1}$)')
    axes[1].set_xlim('2015-01-01', '2016-01-01')
    axes[1].grid(True)

    ax3 = axes[1].twinx()
    lines += ax3.plot(d_emis.overpass, d_emis, 'rD',
                      label='emissions', mfc='none')

    ax3.set_ylim(-32, 32)
    ax3.set_yticks([-20, 0, 20])
    ax3.set_ylabel('wind error (Mt yr$^{-1}$)')

    ax3.legend(lines, [l.get_label() for l in lines], ncol=2)

    plt.tight_layout()
    fig.align_ylabels()

    mb = np.nanmean(d_wind)
    sd = np.nanstd(d_wind)

    print(mb, sd)

    re = scipy.stats.linregress(d_wind, d_emis)

    print(re)






























