

"""
Extracting wind information from 4D wind fields.
"""

from datetime import datetime
import os

import numpy as np
import pandas
import xarray

import smartcarb.config as cfg
import smartcarb.io as io
import smartcarb.misc as misc

import level2
import amrs.misc.chem


def create_pbl_wind_speed_ts(origin, times=None):
    """
    Create a time series of mean wind speeds in the planet boundary layer 50 km
    around the origin.
    """
    data = xarray.Dataset()

    if times is None:
        data['time'] = xarray.DataArray(
            np.arange('2015-01-01 00', '2015-12-26 00', dtype='datetime64[h]'),
            dims='time', attrs={'units': 'UTC'}
        )
    else:
        data['time'] = xarray.DataArray(
            times.values,
            dims='time', attrs={'units': 'UTC'}
        )

    data['wind_speed'] = xarray.DataArray(
        np.full(data['time'].shape, np.nan),
        dims='time', attrs={'long_name': 'wind speed in planet boundary layer',
                            'units': 'm s-1'}
    )
    data['pbl_height'] = xarray.DataArray(
        np.full(data['time'].shape, np.nan),
        dims='time', attrs={'long_name': 'planet boundary layer height',
                            'units': 'm'}
    )


    # constant fields
    path = os.path.join(cfg.DATA_PATH, 'cosmo_2015-01')
    filename = os.path.join(path, 'lffd2015010100c.nc')
    lon, lat, hhl = io.read_fields(filename, ['lon', 'lat', 'HHL'])

    # 50-km radius around origin
    lon_o, lat_o, _ = misc.get_source_origin(origin)
    area = np.sqrt(
        (np.cos(np.deg2rad(lat_o)) *  (lon - lon_o))**2 +
        (lat -lat_o)**2
    ) <= 0.45

    # vertical layers
    levels = hhl[:,area].mean(1)
    layers = 0.5 * (levels[:-1] + levels[1:])

    for i, time in enumerate(data['time']):

        time = pandas.Timestamp(time.values)

        print(i, time)

        # read fields
        if time < np.datetime64('2015-07-01'):
            starttime = datetime(2015,1,1)
        else:
            starttime = datetime(2015,7,1)

        # time-varying fields
        path = os.path.join(cfg.DATA_PATH, starttime.strftime('cosmo_%Y-%m'))
        filename = os.path.join(path, time.strftime('lffd%Y%m%d%H.nc'))

        fields = ['U', 'V', 'HPBL']

        try:
            u, v, pbl = io.read_fields(filename, fields, time=time)
        except FileNotFoundError as e:
            print(e)
            continue

        # get pbl height and wind speed
        pbl_height = pbl[area].mean()

        # wind in area around origin
        u = u[:,area].mean(1)
        v = v[:,area].mean(1)

        wind = np.sqrt(u**2 + v**2)

        # add to dataset
        data['pbl_height'][i] = pbl_height
        data['wind_speed'][i] = wind[layers < pbl_height].mean()

    return data




def plot_wind_profile(tracer, time, show=True):

    import matplotlib.pyplot as plt
    from mpl_toolkits.axisartist.parasite_axes import HostAxes, ParasiteAxes


    time = pandas.Timestamp(time)
    hhl, lon, lat, u, v, c, u10, v10, p, T = read_3d_wind_fields(tracer, time)

    # TODO: read XCO2 tracer and use 0.05 ppm 50-100km 
    mask = c[-1] > 1e-6

    c = amrs.misc.chem.convert_unit(c, 'kg kg-1', 'mg m-3', molar_mass='CO2',
                                    p=p, T=T)


    hhl = hhl[:,mask].mean(1)
    hhl = 0.5 * (hhl[1:] + hhl[:-1]) / 1e3 # km

    c = c[:,mask].mean(1)
    T = T[:,mask].mean(1) - 273.15
    p = p[:,mask].mean(1)
    u = u[:,mask].mean(1)
    v = v[:,mask].mean(1)


    pbl = hhl <= 3.5


    Tmin = np.floor(T[pbl].min() / 5.0) * 5.0
    Tmax = np.ceil(T[pbl].max() / 5.0) * 5.0



    ueff = np.sum(c * np.sqrt(u**2 + v**2)) / np.sum(c)
    u1km = np.mean(np.sqrt(u**2 + v**2)[hhl <= 1.0])


    title = 'u$_{eff}$ = %.1f m s$^{-1}$, u$_{1 km}$ = %.1f m s$^{-1}$' % (ueff, u1km)

    print('ueff:  %5.2f' % ueff)
    print('u1km: %5.2f' % u1km)
    print('diff: %5.2f' % (u1km-ueff))
    print(Tmin, Tmax)

    if show:

        fig = plt.figure(figsize=(3,7))
        host = HostAxes(fig, [0.2, 0.32, 0.73, 0.63])
        par1 = ParasiteAxes(host, sharey=host)
        par2 = ParasiteAxes(host, sharey=host)

        host.parasites.append(par1)
        host.parasites.append(par2)

        fig.add_axes(host)

        par1.axis["bottom"].label.set_visible(True)
        par1.set_ylabel("Velocity")
        offset = (0, -60)
        new_axisline = par1.get_grid_helper().new_fixed_axis
        par1.axis["bottom2"] = new_axisline(loc="bottom", axes=par1,
                                            offset=offset)

        par2.axis["bottom"].label.set_visible(True)
        par2.set_ylabel("Velocity")
        offset = (0, -120)
        new_axisline = par2.get_grid_helper().new_fixed_axis
        par2.axis["bottom3"] = new_axisline(loc="bottom", axes=par2,
                                            offset=offset)


        c1, = host.plot(c[pbl], hhl[pbl], color='orange', label='concentration')
        host.set_ylabel('altitude (km)')
        host.set_ylim(0,3)
        host.set_xlim(-.1, 10.0)
        host.set_xlabel('CO$_2$ concentration (mg m$^{-3}$)')

        c2, = par1.plot(T[pbl], hhl[pbl], color='red',
                        label='temperature')
        par1.set_xlabel('temperature (°C)')

        par1.set_xlim(Tmin, Tmax)

        c3, = par2.plot(np.sqrt(u**2 + v**2)[pbl], hhl[pbl], color='blue',
                        label='wind speed')
        par2.set_xlabel('wind speed (m s$^{-1}$)')
        par2.set_xlim(0,15)

        host.axis["bottom"].label.set_color(c1.get_color())
        par1.axis["bottom2"].label.set_color(c2.get_color())
        par2.axis["bottom3"].label.set_color(c3.get_color())

        host.grid(True, axis='y')
        host.legend(title=str(time)[:10])
        host.set_title(title)






def read_3d_wind_fields(tracer, time):

    if time < datetime(2015,7,1):
        starttime = datetime(2015,1,1)
    else:
        starttime = datetime(2015,7,1)

    # time-varying fields
    path = os.path.join(cfg.DATA_PATH, starttime.strftime('cosmo_%Y-%m'))
    filename = os.path.join(path, time.strftime('lffd%Y%m%d%H.nc'))

    tracer_name = {
        'CO2': 'CO2_BV',
        'NO2': 'NOX_BV',
    }[tracer]
    fields = ['lon', 'lat', 'U', 'V', tracer_name, 'U_10M', 'V_10M', 'HPBL',
              'P', 'T']
    lon, lat, u, v, c, u10, v10, pbl, p, T = io.read_fields(filename, fields, time=time)

    # constant fields
    filename = os.path.join(path, starttime.strftime('lffd%Y%m%d%Hc.nc'))
    hhl = io.read_fields(filename, ['HHL'])[0]

    return hhl, lon, lat, u, v, c, u10, v10, p, T



def compute_effective_wind(h, u, v, c, p, T, curve=None, lon=None, lat=None,
                           origin='Berlin'):
    """
    Compute effective u- and v-wind by weighting with concentration field.

    h: vertical levels
    c: CO2/NO2 mixing ratio
    u: u-wind
    v: v-wind
    """

    if curve is not None: # project wind direction on curve
        x, y = misc.xy2wgs(lon, lat, 'Berlin', inverse=True,
                           use_xarray=False)

        t = curve.get_parameter(x=x, y=y)
        t = curve.compute_tangent(t0=t, norm=True)

        wind = u * t[0] + v * t[1]


    dh = np.diff(h, axis=0)
    c = amrs.misc.chem.convert_unit(c, 'kg kg-1', 'kg m-3', molar_mass='CO2',
                                    p=p, T=T)

    u_eff = np.sum(u * c * dh, axis=0) / np.sum(c * dh, axis=0)
    v_eff = np.sum(v * c * dh, axis=0) / np.sum(c * dh, axis=0)

    u_eff = np.ma.array(u_eff, mask=np.isnan(u_eff))
    v_eff = np.ma.array(v_eff, mask=np.isnan(v_eff))

    if curve is not None:
        wind_eff = np.sum(wind * c * dh, axis=0) / np.sum(c * dh, axis=0)
        wind_eff = np.ma.array(wind_eff, mask=np.isnan(wind_eff))

    else:
        wind_eff = None


    return u_eff, v_eff, wind_eff



def add_wind(data, tracer='CO2', curve=None):

    # read wind fields
    hhl, lon, lat, u, v, c, u10, v10, p, T = read_3d_wind_fields(tracer,
                                                                 data.attrs['time'])
    # effective wind vectors
    u_eff, v_eff, wind_eff = compute_effective_wind(hhl, u, v, c, p, T,
                                                    curve=curve, lon=lon,
                                                    lat=lat)

    # do not compute for these pixels
    mask = ~data['plume_area'].values

    if wind_eff is None:
        fields = level2.sum_satellite_pixels([u_eff, v_eff, u10, v10], data.lonc.values,
                                             data.latc.values, ignore_pixels=mask)
    else:
        fields = level2.sum_satellite_pixels([u_eff, v_eff, u10, v10, wind_eff], data.lonc.values,
                                             data.latc.values, ignore_pixels=mask)

    dims = ('nobs', 'nrows')
    data['u_eff_%s' % tracer] = xarray.DataArray(fields[0], dims=dims, attrs={'units': 'm s-1'})
    data['v_eff_%s' % tracer] = xarray.DataArray(fields[1], dims=dims, attrs={'units': 'm s-1'})
    data['u_10m'] = xarray.DataArray(fields[2], dims=dims, attrs={'units': 'm s-1'})
    data['v_10m'] = xarray.DataArray(fields[3], dims=dims, attrs={'units': 'm s-1'})

    if wind_eff is not None:
        data['wind_eff_%s' % tracer] = xarray.DataArray(fields[4], dims=dims, attrs={'units': 'm s-1'})


    return data

